import 'dart:async';
import 'dart:collection';

import 'package:eyflutter_core/eyflutter_core.dart';
import 'package:eyflutter_player/beans/player_info.dart';
import 'package:eyflutter_player/enums/play_mode.dart';
import 'package:fijkplayer/fijkplayer.dart';

class AudioUtils {
  factory AudioUtils() => _getInstance();

  static AudioUtils get instance => _getInstance();
  static AudioUtils? _instance;

  static AudioUtils _getInstance() {
    return _instance ??= AudioUtils._internal();
  }

  AudioUtils._internal();

  final LinkedHashMap<String, PlayerInfo> _items = LinkedHashMap();
  final Map<String, FijkPlayer> playMap = {};
  ListQueue<PlayerInfo> queue = ListQueue<PlayerInfo>();

  bool _isPreparing = false;
  final Map<int, void Function()> _callMap = {};
  final StreamController<double> _playProgressController = StreamController.broadcast();

  //播放进度更新
  Stream<double> get onPlayProgressUpdate => _playProgressController.stream;

  final StreamController<String> _playChangeController = StreamController.broadcast();

  /// 播放改变监听
  Stream<String> get onPlayChangeUpdate => _playChangeController.stream;

  final StreamController<PlayerInfo> _currentItemChangeController = StreamController.broadcast();

  /// 当前项改变监听
  Stream<PlayerInfo> get onCurrentItemChangeUpdate => _currentItemChangeController.stream;

  /// 播放索引
  int playIndex = 0;
  int _subIndex = 0;

  PlayMode _mode = PlayMode.cycle;

  /// 播放模式
  PlayMode get mode => _mode;

  /// 设置播放模式
  void setPlayMode(PlayMode mode) {
    _mode = mode;
  }

  /// 播放列表
  List<PlayerInfo> get playList => _items.values.toList();

  PlayerInfo _getPlayInfo() {
    if (_items.isEmptyMap() || (playIndex + 1) > _items.length) {
      return PlayerInfo(playerId: "", urlList: []);
    }
    return playList[playIndex];
  }

  /// 获取播放信息
  PlayerInfo get playInfo => _getPlayInfo();

  FijkPlayer? _getPlayer() {
    var playInfo = _getPlayInfo();
    var urlList = playInfo.urlList;
    if (urlList.isEmptyList || _subIndex >= urlList.length) {
      return null;
    }
    var playCode = urlList[_subIndex].hashCode;
    return playMap["${playInfo.playerId}$playCode"];
  }

  /// 获取播放器
  FijkPlayer? get player => _getPlayer();

  FijkPlayer _instancePlayer(String url) {
    return FijkPlayer()..setDataSource(url);
  }

  /// 追加播放列表
  void appendPlayList({required List<PlayerInfo> list}) {
    for (var item in list) {
      if (_items.containsKey(item.playerId)) {
        continue;
      }
      queue.addFirst(item);
    }
    _startLoad();
  }

  /// 追加播放信息
  void appendPlay({required PlayerInfo playerInfo}) {
    if (!_items.containsKey(playerInfo.playerId)) {
      queue.addFirst(playerInfo);
      _startLoad();
    }
  }

  void _startLoad() {
    if (!_isPreparing) {
      _prepare();
    }
  }

  _prepare() {
    if (queue.isEmpty) {
      _isPreparing = false;
      return;
    }
    _isPreparing = true;
    var removeLast = queue.removeLast();
    if (_items.containsKey(removeLast.playerId)) {
      _prepare();
      return;
    }
    for (var element in removeLast.urlList) {
      var player = _instancePlayer(element);
      var playerCode = element.hashCode;
      _callMap[playerCode] = () {
        if (player.state == FijkState.initialized) {
          playMap["${removeLast.playerId}$playerCode"] = player;
          player.removeListener(_callMap[playerCode]!);
          var subIndex = removeLast.urlList.indexOf(element);
          if ((subIndex + 1) == removeLast.urlList.length) {
            _items[removeLast.playerId] = removeLast;
            _isPreparing = false;
            _prepare();
          }
        } else if (player.state == FijkState.error) {
          player.removeListener(_callMap[playerCode]!);
          //资源无效加载下一个
          var subIndex = removeLast.urlList.indexOf(element);
          if ((subIndex + 1) == removeLast.urlList.length) {
            _isPreparing = false;
            _prepare();
          }
        }
      };
      player.addListener(_callMap[playerCode]!);
      player.onCurrentPosUpdate.listen((event) => _playProgressListen(player.value.duration, event));
    }
  }

  void _playProgressListen(Duration total, Duration duration) {
    var currentTime = duration.inMilliseconds.toDouble();
    var totalTime = total.inMilliseconds.toDouble();
    var progress = currentTime / totalTime;
    _playProgressController.add(progress);
    if (progress == 1.0) {
      _switchAudio();
    }
  }

  void _switchAudio() {
    if (_mode == PlayMode.cycle) {
      var _player = player;
      _player?.seekTo(0);
      _player?.pause();
      var _playInfo = playInfo;
      var urlList = _playInfo.urlList;
      if ((_subIndex + 1) < urlList.length) {
        _subIndex++;
      } else {
        if ((playIndex + 1) < _items.length) {
          playIndex++;
          _subIndex = 0;
        } else {
          playIndex = 0;
          _subIndex = 0;
        }
      }
      _player = player;
      _player?.start();
    } else if (_mode == PlayMode.sequence && (playIndex + 1) < _items.length) {
      var _player = player;
      _player?.seekTo(0);
      _player?.pause();
      var _playInfo = playInfo;
      var urlList = _playInfo.urlList;
      if ((_subIndex + 1) < urlList.length) {
        _subIndex++;
      } else {
        playIndex++;
        _subIndex = 0;
      }
      _player = player;
      _player?.start();
    } else if (_mode == PlayMode.singleCycle) {
      var _player = player;
      _player?.seekTo(0);
      _player?.pause();
      var _playInfo = playInfo;
      var urlList = _playInfo.urlList;
      if ((_subIndex + 1) < urlList.length) {
        _subIndex++;
      } else {
        _subIndex = 0;
      }
      _player = player;
      _player?.start();
    }
  }

  /// 开始播放
  void start() {
    var _player = player;
    var total = _player?.value.duration.inMilliseconds ?? 0;
    var milliseconds = _player?.currentPos.inMilliseconds ?? 0;
    if (milliseconds == total) {
      _player?.seekTo(0);
    } else {
      var targetSec = milliseconds - 500;
      if (targetSec > 0) {
        _player?.seekTo(targetSec);
      }
    }
    var _playInfo = playInfo;
    _playChangeController.add(_playInfo.playerId);
    _currentItemChangeController.add(_playInfo);
    _player?.start();
  }

  /// 重置播放
  void reset() {
    var _player = player;
    _player?.seekTo(0);
    _player?.pause();
  }

  /// 暂停播放
  void pause() {
    player?.pause();
  }

  /// 释放资源
  void dispose() {
    playMap.forEach((key, value) {
      value.release();
    });
    _items.clear();
  }

  /// 切换至上一个音频
  void prev() {
    if (playIndex > 0) {
      var _player = player;
      _player?.seekTo(0);
      _player?.pause();
      playIndex--;
      _subIndex = 0;
      _player = player;
      _player?.start();
    }
  }

  /// 切换至下一个视频
  void next() {
    if ((playIndex + 1) < _items.length) {
      var _player = player;
      _player?.seekTo(0);
      _player?.pause();
      playIndex++;
      _subIndex = 0;
      _player = player;
      _player?.start();
    }
  }

  /// 是否第一首
  bool get isFirst {
    if (playIndex == 0 || _items.isEmptyMap()) {
      return true;
    } else {
      return false;
    }
  }

  /// 是否最后一首
  bool get isLast {
    if ((playIndex + 1) == _items.length || _items.isEmptyMap()) {
      return true;
    } else {
      return false;
    }
  }
}
